/*
 * WorldEdit, a Minecraft world manipulation toolkit
 * Copyright (C) sk89q <http://www.sk89q.com>
 * Copyright (C) WorldEdit team and contributors
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package com.sk89q.worldedit.command;

import com.sk89q.worldedit.EditSession;
import com.sk89q.worldedit.LocalSession;
import com.sk89q.worldedit.WorldEdit;
import com.sk89q.worldedit.WorldEditException;
import com.sk89q.worldedit.command.util.CommandPermissions;
import com.sk89q.worldedit.command.util.CommandPermissionsConditionGenerator;
import com.sk89q.worldedit.entity.Player;
import org.enginehub.piston.annotation.Command;
import org.enginehub.piston.annotation.CommandContainer;
import org.enginehub.piston.annotation.param.Arg;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * Commands to undo, redo, and clear history.
 */
@CommandContainer(superTypes = CommandPermissionsConditionGenerator.Registration.class)
public class HistoryCommands {

    private final WorldEdit worldEdit;

    /**
     * Create a new instance.
     *
     * @param worldEdit reference to WorldEdit
     */
    public HistoryCommands(WorldEdit worldEdit) {
        checkNotNull(worldEdit);
        this.worldEdit = worldEdit;
    }

    @Command(
        name = "undo",
        aliases = { "/undo" },
        desc = "Undoes the last action (from history)"
    )
    @CommandPermissions("worldedit.history.undo")
    public void undo(Player player, LocalSession session,
                     @Arg(desc = "Number of undoes to perform", def = "1")
                         int times,
                     @Arg(name = "player", desc = "Undo this player's operations", def = "")
                         String playerName) throws WorldEditException {
        times = Math.max(1, times);
        for (int i = 0; i < times; ++i) {
            LocalSession undoSession = session;
            if (playerName != null) {
                player.checkPermission("worldedit.history.undo.other");
                LocalSession sess = worldEdit.getSessionManager().findByName(playerName);
                if (sess == null) {
                    player.printError("Unable to find session for " + playerName);
                    break;
                }
                undoSession = session;
            }
            EditSession undone = undoSession.undo(undoSession.getBlockBag(player), player);
            if (undone != null) {
                player.print("Undo successful.");
                worldEdit.flushBlockBag(player, undone);
            } else {
                player.printError("Nothing left to undo.");
                break;
            }
        }
    }

    @Command(
        name = "redo",
        aliases = { "/redo" },
        desc = "Redoes the last action (from history)"
    )
    @CommandPermissions("worldedit.history.redo")
    public void redo(Player player, LocalSession session,
                     @Arg(desc = "Number of redoes to perform", def = "1")
                         int times,
                     @Arg(name = "player", desc = "Redo this player's operations", def = "")
                         String playerName) throws WorldEditException {
        times = Math.max(1, times);
        for (int i = 0; i < times; ++i) {
            LocalSession redoSession = session;
            if (playerName != null) {
                player.checkPermission("worldedit.history.redo.other");
                LocalSession sess = worldEdit.getSessionManager().findByName(playerName);
                if (sess == null) {
                    player.printError("Unable to find session for " + playerName);
                    break;
                }
                redoSession = session;
            }
            EditSession redone = redoSession.redo(redoSession.getBlockBag(player), player);
            if (redone != null) {
                player.print("Redo successful.");
                worldEdit.flushBlockBag(player, redone);
            } else {
                player.printError("Nothing left to redo.");
                break;
            }
        }
    }

    @Command(
        name = "clearhistory",
        aliases = { "/clearhistory" },
        desc = "Clear your history"
    )
    @CommandPermissions("worldedit.history.clear")
    public void clearHistory(Player player, LocalSession session) throws WorldEditException {
        session.clearHistory();
        player.print("History cleared.");
    }

}
